'''
AUTHORS:
- Alec Krueger, Texas A&M University, Geochemical and Environmental Research Group, alecmkrueger@tamu.edu.
- Xiao Ge, Texas A&M University, Geochemical and Environmental Research Group, gexiao@tamu.edu.
- Sakib Mahmud, Texas A&M University, Geochemical and Environmental Research Group, sakib@tamu.edu.


Module to ingest and process raw glider data into NetCDF files
'''
from .processor import Processor
from .variable import Variable



import numpy as np
import pandas as pd
import xarray as xr
from pathlib import Path
from attrs import define, field
import datetime
import dbdreader
import shutil
import gsw
import time
import random
import os

from glider_ingest.utils import get_polygon_coords,print_time,find_nth
from glider_ingest.variable import Variable
from glider_ingest.gridder import Gridder
from glider_ingest.dataset_attrs import get_default_variables, get_global_attrs


@define
class Processor:
    """
    A class to process glider data
    """
    # Required attributes
    memory_card_copy_path: Path
    working_dir: Path
    mission_num: str
    # Default attributes
    mission_vars: list[Variable] = field(factory=list)
    glider_ids: dict = field(default={'199': 'Dora', '307': 'Reveille', '308': 'Howdy', '540': 'Stommel', '541': 'Sverdrup', '1148': 'unit_1148'})
    wmo_ids: dict = field(default={'199': 'unknown', '307': '4801938', '308': '4801915', '540': '4801916', '541': '4801924', '1148': '4801915'})
    
    # Optional attributes
    mission_start_date: datetime.datetime = field(default=pd.to_datetime('2010-01-01'))  # Used to slice the data during processing
    mission_end_date: datetime.datetime = field(default=pd.to_datetime(datetime.datetime.today()+datetime.timedelta(days=365)))  # Used to slice the data during processing
    
    # Created attributes
    df: pd.DataFrame = field(default=None)
    ds: xr.Dataset = field(default=None)
    
    # Private backing fields
    _glider_id: str = field(default=None)
    _glider_name: str = field(default=None)
    _wmo_id: str = field(default=None)
    _mission_year: str = field(default=None)
    _mission_title: str = field(default=None)
    _mission_folder_name: str = field(default=None)
    _mission_folder_path: Path = field(default=None)
    _netcdf_filename: str = field(default=None)
    _netcdf_output_path: Path = field(default=None)

    @property
    def glider_id(self):
        """Get the glider ID."""
        if self._glider_id is None:
            self._glider_id = self._get_glider_id()
        return self._glider_id

    @property
    def glider_name(self):
        """Get the glider name."""
        if self._glider_name is None:
            self._glider_name = self.glider_ids[self.glider_id]
        return self._glider_name

    @property
    def wmo_id(self):
        """Get the WMO ID."""
        if self._wmo_id is None:
            self._wmo_id = self.wmo_ids[self.glider_id]
        return self._wmo_id
    
    @property
    def mission_year(self):
        """Get the mission year."""
        if self._mission_year is None:
            self._mission_year = self._get_mission_year()
        return self._mission_year

    @property
    def mission_title(self):
        """Get the mission title."""
        if self._mission_title is None:
            self._mission_title = f'Mission {self.mission_num}'
        return self._mission_title

    @property
    def mission_folder_name(self):
        """Get the mission folder name."""
        if self._mission_folder_name is None:
            self._mission_folder_name = self.mission_title.replace(' ', '_')
        return self._mission_folder_name

    @property
    def mission_folder_path(self):
        """Get the mission folder path."""
        if self._mission_folder_path is None:
            self._mission_folder_path = self.working_dir.joinpath(self.mission_folder_name)
        return self._mission_folder_path

    @property
    def netcdf_filename(self):
        """Get the NetCDF filename."""
        if self._netcdf_filename is None:
            self._netcdf_filename = f'M{self.mission_num}_{self.mission_year}_{self.glider_id}.nc'
        return self._netcdf_filename

    @property
    def netcdf_output_path(self):
        """Get the NetCDF path."""
        if self._netcdf_output_path is None:
            self._netcdf_output_path = self.mission_folder_path.joinpath(f'{self.netcdf_filename}')
        return self._netcdf_output_path

    
    def __attrs_post_init__(self):
        """
        Post init method to add default variables to the mission_vars list
        """
        self.add_mission_vars(get_default_variables())
        
        
    def _check_mission_var_duplicates(self):
        """
        Check for duplicate variables in the mission_vars list
        """
        # Get the variable data_source_names
        var_names = self._get_mission_variable_short_names()
        if len(set(var_names)) != len(var_names):
            print('Duplicate variables in mission_vars list')
        
    def add_mission_vars(self, mission_vars: list[Variable]):
        """
        Add a variable to the mission_vars list
        """
        if all(isinstance(var, str) for var in mission_vars):
            mission_vars = [Variable(data_source_name=var) for var in mission_vars]
        self.mission_vars.extend(mission_vars)
        self._check_mission_var_duplicates()
        
   
    def _copy_files(self):
        """
        Copy only LOGS and STATE/CACHE folders from memory card copy to working directory
        """
        original_loc = self.memory_card_copy_path
        new_loc = self.mission_folder_path
        
        # Define patterns to include
        include_patterns = ['**/LOGS', '**/logs', '**/STATE/CACHE', '**/state/cache']
        
        for pattern in include_patterns:
            for source_path in original_loc.glob(pattern):
                # Create relative path to maintain directory structure
                relative_path = source_path.relative_to(original_loc)
                destination_path = new_loc / relative_path
                
                # Create parent directories if they don't exist
                destination_path.parent.mkdir(parents=True, exist_ok=True)
                
                # Copy directory
                if source_path.is_dir():
                    shutil.copytree(source_path, destination_path, dirs_exist_ok=True)


    def _get_files_by_extension(self,directory_path: Path, extensions: list[str], as_string: bool = False) -> list:
        """
        Get files from a directory with specified extensions.

        Args:
            directory_path (Path): Directory to search for files
            extensions (list[str]): List of file extensions to match (e.g. ['.dbd', '.DBD'])
            as_string (bool): Whether to return paths as strings

        Returns:
            list: List of matching files as Path objects or strings
        """
        files = [p for p in directory_path.rglob('*') if p.suffix in extensions]
        if as_string:
            files = [str(p) for p in files]
        return files
        
    def _get_cache_files(self,as_string:bool=False):
        """
        Get the cache files from the memory card copy
        """
        extensions = ['.cac','.CAC']
        directory_path = self.mission_folder_path
        cac_files = self._get_files_by_extension(directory_path=directory_path,extensions=extensions,as_string=as_string)
        return cac_files

    def _get_cache_files_path(self):
        """
        Get the cache file path from the memory card copy
        """
        return self.mission_folder_path.joinpath('cache')

    def _copy_cache_files(self):
        """
        Move the cache files to the working directory
        """
        cache_files = self._get_cache_files()
        new_cache_loc = self._get_cache_files_path()
        if not new_cache_loc.exists():
            new_cache_loc.mkdir()
        for cache_file in cache_files:
            # check if cache file already exists
            if new_cache_loc.joinpath(cache_file.name).exists():
                continue
            # copy cache file
            new_cache_filename = new_cache_loc.joinpath(cache_file.name)
            print(f'Coping {cache_file} to {new_cache_loc}')
            shutil.copy2(cache_file, new_cache_filename)
            
    def _get_dbd_files(self,as_string=False):
        """
        Get the dbd files from the memory card copy
        """
        directory_path = self.mission_folder_path
        extensions = ['.dbd','.DBD','.ebd','.EBD']
        dbd_files = self._get_files_by_extension(directory_path=directory_path,extensions=extensions,as_string=as_string)
        return dbd_files
        
    def _read_dbd(self):
        """
        Read the files from the memory card copy
        """
        self._copy_files()
        time.sleep(2)
        self._copy_cache_files()
        
        filenames = self._get_dbd_files(as_string=True)
        dbd = dbdreader.MultiDBD(filenames=filenames,cacheDir=self._get_cache_files_path())
        return dbd
    
    def _get_dbd_variables(self,dbd) -> list:
        """
        Get the dbd variables from the files. Returns both sci and eng if both are true.
        
        If sci_vars and eng_vars are both False, raise a ValueError
        
        
        Parameters
        ----------
        sci_vars : bool, optional
        Whether to include science variables, default: True
        eng_vars : bool, optional
        Whether to include engineering variables, default: True
        
        Returns
        -------
        list
        
        Raises
        ------
        ValueError
        If sci_vars and eng_vars are both False
        """
        sci_dbd_vars = dbd.parameterNames['sci']
        eng_dbd_vars = dbd.parameterNames['eng']
        all_dbd_vars = sci_dbd_vars + eng_dbd_vars
        return all_dbd_vars
    
    def _get_mission_variables(self,filter_out_none=False):
        """
        Get the mission variables from the mission_vars list. Filter out None data_source_name values if desired.
        """
        if filter_out_none:
            return [var for var in self.mission_vars if var.data_source_name is not None]
        else:
            return self.mission_vars
    
    def _get_mission_variable_short_names(self,filter_out_none=False):
        """
        Get the mission variable data source names from the mission_vars list
        """
        return [var.short_name for var in self._get_mission_variables(filter_out_none=filter_out_none)]
    
    def _get_mission_variable_data_source_names(self,filter_out_none=False):
        """
        Get the mission variable data source names from the mission_vars list
        """
        return [var.data_source_name for var in self._get_mission_variables(filter_out_none=filter_out_none)]
    
    def _check_default_variables(self,dbd,variables_to_get:list):
        """
        Check that the default variables are in the dbd variables
        """
        dbd_vars = self._get_dbd_variables(dbd=dbd)
        missing_vars = [var for var in variables_to_get if var not in dbd_vars]
        if missing_vars:
            print(f'The following variables are missing from the dbd files: {missing_vars}')
        
    def _get_sci_files(self):
        """
        Get the sci files from the memory card copy
        """
        directory_path = self.mission_folder_path
        extensions = ['.dbd','.DBD']
        sci_files = self._get_files_by_extension(directory_path=directory_path,extensions=extensions,as_string=True)
        return sci_files
        
    def _get_random_sci_file(self):
        """
        Get a random sci file from the mission folder
        """
        sci_files = self._get_sci_files()
        random_file = random.choice(sci_files)
        # Pick a new file if the file is empty
        while os.stat(random_file).st_size == 0:
            random_file = random.choice(sci_files)
        return random_file
        
    def _get_full_filename(self):
        """
        Get the full filename from the file
        Args:
            file (str): Path to the file to read.

        Returns:
            str: The extracted full filename, or None if not found.
        """
        file = self._get_random_sci_file()
        with open(file, errors="ignore") as fp:
            for line in fp:
                if 'full_filename' in line.strip():
                    return line.replace('full_filename:', '').strip()
        return None
    
    def _get_mission_year(self):
        """
        Get the mission year from the filename.

        Extracts and validates the mission year from the filename, converting between
        mission names and IDs as needed using the mission_ids mapping.

        Returns
        -------
        str
            The validated mission year
        """
        full_filename = self._get_full_filename()
        mission_year = full_filename[full_filename.find('-') + 1: find_nth(full_filename, '-', 2)].strip()
        return mission_year
        
    def _get_glider_id(self):
        """
        Get the glider id from the filename.
        
        Extracts and validates the glider identifier from the filename, converting between
        glider names and IDs as needed using the glider_ids mapping.
        
        Returns
        -------
        str
            The validated glider ID
        """
        full_filename = self._get_full_filename()
        glider_identifier = full_filename.split('-')[0].replace('unit_', '').strip()
        
        # Create reverse mapping from names to IDs
        inverted_glider_ids = {v: k for k, v in self.glider_ids.items()}
        
        # Check if identifier is a valid ID
        if glider_identifier in self.glider_ids:
            return glider_identifier
            
        # Check if identifier is a valid name
        if glider_identifier in inverted_glider_ids:
            return inverted_glider_ids[glider_identifier]
            
        valid_options = list(self.glider_ids.keys()) + list(self.glider_ids.values())
        print(f'Invalid glider identifier: {glider_identifier}. Must be one of: {valid_options}')
        return None
    
    def _get_dbd_data(self):
        dbd = self._read_dbd()
        variables_to_get = self._get_mission_variable_data_source_names(filter_out_none=True)
        self._check_default_variables(dbd,variables_to_get)
        data = dbd.get_sync(*variables_to_get)
        dbd.close()
        return data
    
    def _format_time(self,df:pd.DataFrame):
        # Convert time to datetime format and filter valid dates
        df['time'] = pd.to_datetime(df['time'],unit='s', errors='coerce')
        df = df.dropna(how='all')
        valid_dates_mask = (df['time'] >= self.mission_start_date) & \
                           (df['time'] <= self.mission_end_date)
        df = df.loc[valid_dates_mask]
        return df
        
    def _calculate_vars(self,df):
        # Perform variable conversions and calculations
        df['m_pressure'] *= 10  # Convert pressure from db to dbar
        df['sci_water_pressure'] *= 10  # Convert pressure from db to dbar
        df['salinity'] = gsw.SP_from_C(df['sci_water_cond'] * 10, df['sci_water_temp'], df['sci_water_pressure'])
        CT = gsw.CT_from_t(df['salinity'], df['sci_water_temp'], df['sci_water_pressure'])
        df['density'] = gsw.rho_t_exact(df['salinity'], CT, df['sci_water_pressure'])
        return df
    
    def _update_dataframe_columns(self,df):
        """
        Update the dataframe columns with the mission variables.
        Adjusting the current column names, which are data source names, to their short_name values.
        """
        column_map = {value.data_source_name: value.short_name for value in self.mission_vars}
        df = df.rename(columns=column_map)
        return df
    
    def _convert_dbd_to_dataframe(self):
        """
        Get the dbd data as a dataframe
        """
        data = self._get_dbd_data()
        df = pd.DataFrame(data).T
        new_column_names = ['time']
        new_column_names.extend(self._get_mission_variable_data_source_names(filter_out_none=True))
        if len(df.columns) != len(new_column_names):
            print(f'The number of columns in the dataframe does not match the number of mission variables, {df.columns} vs {new_column_names}')
        # Add names to the dataframe columns
        df.columns = new_column_names
        # Format time
        df = self._format_time(df)
        # Calculate variables
        df = self._calculate_vars(df)
        # Set time as index
        df = df.set_index('time')
        df = self._update_dataframe_columns(df)
        self.df = df
        return self.df
    
    def _convert_df_to_ds(self):
        df = self._convert_dbd_to_dataframe()
        ds = xr.Dataset.from_dataframe(df)
        self.ds = ds
        return self.ds
    
    def _get_longitude(self):
        return self.ds.longitude.values
    
    def _get_latitude(self):
        return self.ds.latitude.values
    
    def _get_depth(self):
        return self.ds.depth.values
    
    def _get_time(self):
        return self.ds.time.values

    def _add_global_attrs(self):
        from glider_ingest.dataset_attrs import get_global_attrs
        global_attrs = get_global_attrs(wmo_id = self.wmo_id,mission_title=self.mission_title,
                                        longitude=self._get_longitude(),latitude=self._get_latitude(),
                                        depth=self._get_depth(),time=self._get_time())
        
        self.ds.attrs = global_attrs                                        
        
    def _add_variable_attrs(self):
        for var in self.mission_vars:
            self.ds[var.short_name].attrs = var.to_dict()
    
    def _add_attrs(self):
        # Add global attributes
        self._add_global_attrs()
        # Add variable attributes
        self._add_variable_attrs()
        
    def _add_gridded_data(self):
        '''Add gridded data to the dataset, must be called after adding attrs'''
        ds_gridded = Gridder(self.ds).create_gridded_dataset()
        self.ds.update(ds_gridded)

    def process(self,return_ds=True):
        self._convert_df_to_ds()
        self._add_attrs()
        self._add_gridded_data()
        if return_ds:
            return self.ds
        
    def save(self,save_path=None):
        if self.ds is None:
            self.process()
        if save_path is None:
            save_path = self.netcdf_output_path
        self.ds.to_netcdf(save_path)
        return self.ds
        
    
# Test the processor class

memory_card_copy_path = Path('C:/Users/alecmkrueger/Documents/GERG/GERG_GitHub/GERG-Glider/Code/Packages/glider_ingest/src/tests/test_data/memory_card_copy')
working_dir = Path('C:/Users/alecmkrueger/Documents/GERG/GERG_GitHub/GERG-Glider/Code/Packages/glider_ingest/src/tests/test_data/working_dir')

processor = Processor(memory_card_copy_path=memory_card_copy_path,working_dir=working_dir,mission_num='46',mission_start_date='2024-06-28 02:58:28.873474121')

ds = processor.save()
ds


"""
Module containing utility functions for the package.
"""

# Import Packages
import numpy as np
import xarray as xr
import datetime
from functools import wraps
from time import time
import inspect

def print_time(message: str) -> None:
    """
    Print a message with the current time appended.

    Parameters
    ----------
    message : str
        The message to print.

    Notes
    -----
    The current time is formatted as 'HH:MM:SS'.
    """
    # Get current time
    current_time = datetime.datetime.today().strftime('%H:%M:%S')
    # Combine the message with the time
    whole_message = f'{message}: {current_time}'
    # Print the final message
    print(whole_message)
    

def f_print(*args, return_string=False):
    # Get the current frame and extract the calling frame
    frame = inspect.currentframe().f_back
    # Combine locals and globals for name lookup
    all_vars = {**frame.f_globals, **frame.f_locals}
    
    results = []
    for var in args:
        # Find all variable names that match the value of `var`
        var_names = [name for name, val in all_vars.items() if val is var]
        if var_names:
            # Use the first matching variable name
            results.append(f"{var_names[0]} = {var}")
        else:
            results.append(f"Could not determine variable name for value: {var}")
    
    # Return or print the result
    if return_string:
        return ", ".join(results)
    else:
        print(", ".join(results))

def timing(f):
    """Time a function.

    Args:
        f (function): function to time

    Returns:
        wrapper: prints the time it took to run the function
    """
    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        print('func:%r args:[%r, %r] took: %2.4f sec' % \
          (f.__name__, args, kw, te-ts))
        return result
    return wrap

def find_nth(haystack: str, needle: str, n: int) -> int:
    """
    Find the nth occurrence of a substring in a string.

    Parameters
    ----------
    haystack : str
        The string to search in.
    needle : str
        The substring to find.
    n : int
        The occurrence number of the substring to find.

    Returns
    -------
    int
        The index of the nth occurrence of the substring, or -1 if not found.
    """
    # Start at the first occurrence of the substring
    start = haystack.find(needle)
    while start >= 0 and n > 1:
        # Find the next occurrence
        start = haystack.find(needle, start + len(needle))
        n -= 1
    return start

def invert_dict(dict: dict) -> dict:
    """
    Invert the keys and values of a dictionary.

    Parameters
    ----------
    dict : dict
        The dictionary to invert.

    Returns
    -------
    dict
        A new dictionary with keys and values swapped.
    """
    # Create a new dictionary with inverted key-value pairs
    return {value: key for key, value in dict.items()}

def get_polygon_bounds(longitude:np.ndarray,latitude:np.ndarray) -> list:
    """
    Generate polygon coordinates for the dataset's global attributes.
    """
    # Get the maximum latitude below 29.5
    lat_max = np.nanmax(latitude[np.where(latitude < 29.5)])
    # Get the minimum latitude below 29.5
    lat_min = np.nanmin(latitude[np.where(latitude < 29.5)])
    # Get the maximum longitude
    lon_max = np.nanmax(longitude)
    # Get the minimum longitude
    lon_min = np.nanmin(longitude)
    return [lat_max, lat_min, lon_max, lon_min]

def get_polygon_coords(longitude:np.ndarray,latitude:np.ndarray,lat_max:float, lat_min:float, lon_max:float, lon_min:float) -> str:
    """
    Generate polygon coordinates for the dataset's global attributes.

    Parameters
    ----------
    ds_mission : xarray.Dataset
        The mission dataset containing latitude and longitude values.

    Returns
    -------
    str
        A string representation of the polygon in Well-Known Text (WKT) format.

    Notes
    -----
    The polygon is constructed based on the northmost, eastmost, southmost, 
    and westmost points where latitude is below 29.5.
    """

    # lat_max, lat_min, lon_max, lon_min = get_polygon_bounds(longitude,latitude)

    # Construct polygon points
    polygon_1 = f"{lat_max} {longitude[np.where(latitude == lat_max)[0][0]]}"  # Northmost
    polygon_2 = f"{latitude[np.where(longitude == lon_max)[0][0]]} {lon_max}"  # Eastmost
    polygon_3 = f"{lat_min} {longitude[np.where(latitude == lat_min)[0][0]]}"  # Southmost
    polygon_4 = f"{latitude[np.where(longitude == lon_min)[0][0]]} {lon_min}"  # Westmost
    polygon_5 = polygon_1  # Close the polygon

    # Combine points into WKT polygon format
    return f"POLYGON (({polygon_1}, {polygon_2}, {polygon_3}, {polygon_4}, {polygon_5}))"

def get_wmo_id(glider_id: str) -> str:
    """
    Extract the WMO ID from a glider ID.
    """
    if isinstance(glider_id, int):
        glider_id = str(glider_id)
    glider_ids = {'199': 'Dora', '307': 'Reveille', '308': 'Howdy', '540': 'Stommel', '541': 'Sverdrup', '1148': 'unit_1148'}
    wmo_ids = {'199': 'unknown', '307': '4801938', '308': '4801915', '540': '4801916', '541': '4801924', '1148': '4801915'}
    return wmo_ids[glider_id]



from glider_ingest.variable import Variable

def generate_variables(interval_h, interval_p):
    g_temperature = Variable(
        long_name='Gridded Temperature',
        short_name='g_temperature',
        observation_type='calculated',
        source='temperature from sci_water_temp',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='sea_water_temperature',
        units='Celsius',
        valid_max=40.0,
        valid_min=-5.0
    )

    g_salinity = Variable(
        long_name='Gridded Salinity',
        short_name='g_salinity',
        observation_type='calculated',
        source='salinity from sci_water_sal',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='sea_water_practical_salinity',
        units='1',
        valid_max=40.0,
        valid_min=0.0
    )

    g_conductivity = Variable(
        long_name='Gridded Conductivity',
        short_name='g_conductivity',
        observation_type='calculated',
        source='conductivity from sci_water_cond',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='sea_water_electrical_conductivity',
        units='S m-1',
        valid_max=10.0,
        valid_min=0.0
    )

    g_density = Variable(
        long_name='Gridded Density',
        short_name='g_density',
        observation_type='calculated',
        source='density from sci_water_dens',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='sea_water_density',
        units='kg m-3',
        valid_max=1040.0,
        valid_min=1015.0
    )

    g_turbidity = Variable(
        long_name='Gridded Turbidity',
        short_name='g_turbidity',
        observation_type='calculated',
        source='turbidity from sci_flbbcd_bb_units',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='sea_water_turbidity',
        units='1',
        valid_max=1.0,
        valid_min=0.0
    )

    g_cdom = Variable(
        long_name='Gridded CDOM',
        short_name='g_cdom',
        observation_type='calculated',
        source='cdom from sci_flbbcd_cdom_units',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='concentration_of_colored_dissolved_organic_matter_in_sea_water',
        units='ppb',
        valid_max=50.0,
        valid_min=0.0
    )

    g_chlorophyll = Variable(
        long_name='Gridded Chlorophyll_a',
        short_name='g_chlorophyll',
        observation_type='calculated',
        source='chlorophyll from sci_flbbcd_chlor_units',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='mass_concentration_of_chlorophyll_a_in_sea_water',
        units='\u03BCg/L',
        valid_max=10.0,
        valid_min=0.0
    )

    g_oxygen = Variable(
        long_name='Gridded Oxygen',
        short_name='g_oxygen',
        observation_type='calculated',
        source='oxygen from sci_oxy4_oxygen',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='moles_of_oxygen_per_unit_mass_in_sea_water',
        units='\u03BCmol/kg',
        valid_max=500.0,
        valid_min=0.0
    )

    g_hc = Variable(
        long_name='Gridded Heat Content',
        short_name='g_hc',
        observation_type='calculated',
        source='g_temp',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='sea_water_heat_content_for_all_grids',
        units='kJ/cm^2',
        valid_max=10.0,
        valid_min=0.0
    )

    g_phc = Variable(
        long_name='Gridded Potential Heat Content',
        short_name='g_phc',
        observation_type='calculated',
        source='g_temp',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='sea_water_heat_content_for_grids_above_26_C',
        units='kJ/cm^2',
        valid_max=10.0,
        valid_min=0.0
    )

    g_sp = Variable(
        long_name='Gridded Spiciness',
        short_name='g_sp',
        observation_type='calculated',
        source='g_temp',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='spiciness_from_absolute_salinity_and_conservative_temperature_at_0dbar',
        units='kg/m^3',
        valid_max=10.0,
        valid_min=0.0
    )

    g_depth = Variable(
        long_name='Gridded Depth',
        short_name='g_depth',
        observation_type='calculated',
        source='g_pres',
        resolution=str(interval_h)+'hour and '+str(interval_p)+'dbar',
        standard_name='sea_water_depth',
        units='m',
        valid_max=1000.0,
        valid_min=0.0
    )

    # Create a list of variables that we initilized above
    variables = [g_temperature, g_salinity, g_conductivity, g_density, g_turbidity, 
                 g_cdom, g_chlorophyll, g_oxygen, g_hc, g_phc, g_sp, g_depth]
    # Create a dictionary of variables that we initilized above
    attrs_dict = {value.short_name:value for value in variables} 
    return attrs_dict



'''
Module containing the Gridder class.
'''
from attrs import define, field
import numpy as np
import pandas as pd
import xarray as xr
import gsw

@define
class Gridder:
    '''
    Class to create and calculate a gridded dataset from a mission dataset.

    This class provides methods for processing oceanographic data, creating time and pressure grids,
    interpolating data onto those grids, and adding metadata attributes to the gridded dataset.
    
    Depends on the dataset having attributes
    ----------------------------------------

    Attributes:
        ds_mission (xr.Dataset): The input mission dataset to process.
        interval_h (int | float): Time interval (in hours) for gridding.
        interval_p (int | float): Pressure interval (in decibars) for gridding.

    Internal Attributes (initialized later):
        ds (xr.Dataset): A copy of the mission dataset with NaN pressures removed.
        variable_names (list): List of variable names in the dataset.
        time, pres (np.ndarray): Arrays of time and pressure values.
        lat, lon (np.ndarray): Mean latitude and longitude of the dataset.
        grid_pres, grid_time (np.ndarray): Pressure and time grids for interpolation.
        data_arrays (dict): Dictionary of initialized gridded variables.
    '''

    ds_mission: xr.Dataset
    interval_h: int | float = field(default=1)  # Time interval for gridding in hours.
    interval_p: int | float = field(default=0.1)  # Pressure interval for gridding in decibars.

    # Attributes initialized post-construction.
    ds: xr.Dataset = field(init=False)
    ds_gridded: xr.Dataset = field(init=False)
    variable_names: list = field(init=False)
    time: np.ndarray = field(init=False)
    pres: np.ndarray = field(init=False)
    lat: np.ndarray = field(init=False)
    lon: np.ndarray = field(init=False)
    xx: np.ndarray = field(init=False)
    yy: np.ndarray = field(init=False)
    int_time: np.ndarray = field(init=False)
    int_pres: np.ndarray = field(init=False)
    data_arrays: dict = field(init=False)
    grid_pres: np.ndarray = field(init=False)
    grid_time: np.ndarray = field(init=False)

    def __attrs_post_init__(self):
        '''
        Initializes the Gridder class by copying the mission dataset, filtering valid pressures,
        extracting dataset dimensions, and initializing the time-pressure grid.
        '''
        self.ds = self.ds_mission.copy()
        
        # Identify indexes of valid (non-NaN) pressure values.
        tloc_idx = np.where(~np.isnan(self.ds['pressure']))[0]
        
        # Select times corresponding to valid pressures.
        self.ds = self.ds.isel(time=tloc_idx)
        
        # Extract variable names and time/pressure values.
        self.variable_names = list(self.ds.data_vars.keys())
        self.time = self.ds.time.values
        self.check_len(self.time, 1)  # Ensure there is sufficient data to grid.
        self.pres = self.ds.pressure.values

        # Calculate mean latitude and longitude.
        self.lon = np.nanmean(self.ds_mission.longitude.values)
        self.lat = np.nanmean(self.ds_mission.latitude.values)

        # Initialize the time-pressure grid.
        self.initalize_grid()

    def check_len(self, values, expected_length):
        '''
        Ensures that the length of the input array is greater than the expected length.

        Args:
            values (list | np.ndarray): Input array to check.
            expected_length (int): Minimum required length.

        Raises:
            ValueError: If the length of `values` is less than or equal to `expected_length`.
        '''
        if len(values) <= expected_length:
            raise ValueError(f'Not enough values to grid {values}')
        
    def initalize_grid(self):
        '''
        Creates a time-pressure grid for interpolation.

        This method calculates evenly spaced time intervals based on the `interval_h` attribute
        and pressure intervals based on the `interval_p` attribute. The resulting grids are stored
        as internal attributes for further processing.
        '''
        # Define the start and end times rounded to the nearest interval.
        start_hour = int(pd.to_datetime(self.time[0]).hour / self.interval_h) * self.interval_h
        end_hour = int(pd.to_datetime(self.time[-1]).hour / self.interval_h) * self.interval_h
        start_time = pd.to_datetime(self.time[0]).replace(hour=start_hour, minute=0, second=0)
        end_time = pd.to_datetime(self.time[-1]).replace(hour=end_hour, minute=0, second=0)

        # Generate an array of evenly spaced time intervals.
        self.int_time = np.arange(start_time, end_time + np.timedelta64(self.interval_h, 'h'),
                                  np.timedelta64(self.interval_h, 'h')).astype('datetime64[ns]')

        # Create evenly spaced pressure intervals.
        start_pres = 0  # Start pressure in dbar.
        end_pres = np.nanmax(self.pres)  # Maximum pressure in dataset.
        self.int_pres = np.arange(start_pres, end_pres, self.interval_p)

        # Generate the pressure-time grid using a meshgrid.
        self.grid_pres, self.grid_time = np.meshgrid(self.int_pres, self.int_time[1:])
        self.xx, self.yy = np.shape(self.grid_pres)  # Dimensions of the grid.

        # Initialize variables for grid interpolation.
        var_names = [
            f'int_{varname}'
            for varname in self.variable_names
            if self.ds[varname].attrs.get('to_grid') in [True, 'True']
        ]

        # Initialize data arrays with NaN values
        self.data_arrays = {
            var: np.full((self.xx, self.yy), np.nan) 
            for var in var_names
        }


    def add_attrs(self):
        '''
        Adds descriptive metadata attributes to the gridded dataset variables.

        This method assigns long names, units, valid ranges, and other metadata to the
        gridded dataset variables for better interpretation and standardization.
        '''
        from glider_ingest.gridded_attrs import generate_variables
        variables = generate_variables(self.interval_h,self.interval_p)
        for var_short_name,variable in variables.items():
            if var_short_name in self.ds_gridded.data_vars.keys():
                self.ds_gridded[var_short_name].attrs = variable.to_dict()

    def _process_time_slice(self, tds):
        """
        Process a single time slice of data.
        
        Steps:
            - Sort data by pressure
            - Convert time coordinates to datetime64
            - Set time values to pressure values
        """
        if len(tds.time) == 0:
            return tds
            
        tds = tds.sortby('pressure')
        tds = tds.assign_coords(time=('time', tds.time.values.astype('datetime64[ns]')))
        tds['time'] = tds['pressure'].values
        return self._handle_pressure_duplicates(tds)

    def _handle_pressure_duplicates(self, tds):
        """
        Handle duplicate pressure values by adding tiny offsets.
        
        Steps:
            - Identify duplicate pressure values
            - Add small incremental offsets to make values unique
            - Update time values to match new pressure values
        """
        unique_pressures, indices, counts = np.unique(tds['pressure'], return_index=True, return_counts=True)
        duplicates = unique_pressures[counts > 1]
        
        for pressure in duplicates:
            indices = np.where(tds['pressure'] == pressure)[0]
            for idx in indices:
                tds['pressure'][idx] = pressure + 0.000000000001 * idx
        
        tds['time'] = tds['pressure'].values
        return tds

    def _interpolate_variables(self):
        """
        Interpolate variables to fixed pressure grid.
        
        Steps:
            - Select and process time slices
            - Interpolate each variable onto the fixed pressure grid
        """
        for ttt in range(self.xx):
            tds = self.ds.sel(time=slice(str(self.int_time[ttt]), str(self.int_time[ttt+1]))).copy()
            tds = self._process_time_slice(tds)
            
            for data_array_key, value in self.data_arrays.items():
                tds_key = data_array_key.replace('int_', '')
                self.data_arrays[data_array_key][ttt,:] = tds[tds_key].interp(time=self.int_pres)
        
    def _calculate_derived_quantities(self):
        """
        Calculate derived oceanographic quantities.
        
        Computed quantities:
            - Absolute salinity, conservative temperature, and potential temperature
            - Specific heat capacity, spiciness, and depth
        Derived quantities:
            - Heat content (HC): :math:`\\Delta Z \\cdot C_p \\cdot T \\cdot \\rho`
            - Potential heat content (PHC): :math:`\\Delta Z \\cdot C_p \\cdot (T - 26) \\cdot \\rho`, where values < 0 are set to NaN
        """
        sa = gsw.SA_from_SP(self.data_arrays['int_salinity'], self.grid_pres, self.lon, self.lat)
        pt = gsw.pt0_from_t(sa, self.data_arrays['int_temperature'], self.grid_pres)
        ct = gsw.CT_from_pt(sa, pt)
        cp = gsw.cp_t_exact(sa, self.data_arrays['int_temperature'], self.grid_pres) * 0.001
        dep = gsw.z_from_p(self.grid_pres, self.lat, geo_strf_dyn_height=0, sea_surface_geopotential=0)
        spc = gsw.spiciness0(sa, ct)
        
        dz = self.interval_p
        hc = dz * cp * self.data_arrays['int_temperature'] * self.data_arrays['int_density']
        phc = dz * cp * (self.data_arrays['int_temperature'] - 26) * self.data_arrays['int_density']
        phc[phc < 0] = np.nan
        
        return hc, phc, spc, dep

    def _create_output_dataset(self, hc, phc, spc, dep):
        """
        Create the final xarray Dataset with all variables.
        
        Output variables:
            - Gridded variables with 'g_' prefix
            - g_hc: Heat content in kJ cm^{-2}
            - g_phc: Potential heat content in kJ cm^{-2}
            - g_sp: Spiciness
            - g_depth: Depth in meters
        """
        self.ds_gridded = xr.Dataset()
        
        for data_array_key, value in self.data_arrays.items():
            base_key = data_array_key.replace('int_', '')
            if base_key in self.variable_names:
                # print('adding gridded data to xarray', base_key)
                gridded_var = f'g_{base_key}'
                self.ds_gridded[gridded_var] = xr.DataArray(
                    value,
                    [('g_time', self.int_time[1:]), ('g_pres', self.int_pres)]
                )
        
        derived_vars = {
            'g_hc': hc * 10**-4,
            'g_phc': phc * 10**-4,
            'g_sp': spc,
            'g_depth': dep
        }
        
        for var_name, data in derived_vars.items():
            self.ds_gridded[var_name] = xr.DataArray(
                data,
                [('g_time', self.int_time[1:]), ('g_pres', self.int_pres)]
            )

    def create_gridded_dataset(self) -> xr.Dataset:
        """
        Process and interpolate time-sliced data to create a gridded dataset.
        
        This method orchestrates the complete gridding process by:
            1. Interpolating variables onto a fixed pressure grid
            2. Computing derived oceanographic quantities
            3. Creating the final dataset with standardized dimensions
            4. Adding metadata attributes
        
        Note:
            Requires the `gsw` library for oceanographic calculations and assumes
            that `self.data_arrays` and `self.int_time` are properly initialized.
        """
        self._interpolate_variables()
        hc, phc, spc, dep = self._calculate_derived_quantities()
        self._create_output_dataset(hc, phc, spc, dep)
        self.add_attrs()
        return self.ds_gridded



from attrs import define, field, asdict
import pandas as pd

from glider_ingest.utils import get_wmo_id

@define
class Variable:
    """
    A class to represent a variable in a glider mission dataset.
    """
    # Required attributes
    data_source_name: str|None = field(default=None)  # Name of the variable in the data source
    short_name: str|None = field(default=None)  # Name of the variable in the dataset, if it is changed from the data source
    # Optional attributes
    accuracy: float|None = field(default=None)
    ancillary_variables: str|None = field(default=None)
    instrument: str|None = field(default=None)
    long_name: str|None = field(default=None)  
    observation_type: str|None = field(default=None)
    resolution: str|None|float = field(default=None)
    axis: str|None = field(default=None)
    bytes: str|None|int = field(default=None)
    comment: str|None = field(default=None)
    observation_type: str|None = field(default=None)
    platform: str|None = field(default=None)
    positive: str|None = field(default=None)
    precision: str|None|float = field(default=None)
    reference_datum: str|None = field(default=None)
    source_sensor: str|None = field(default=None)
    standard_name: str|None = field(default=None)
    units: str|None = field(default=None)
    valid_max: str|None|float = field(default=None)
    valid_min: str|None|float = field(default=None)
    update_time: str|None = field(default=None)
    coordinate_reference_frame: str|None = field(default=None)
    source: str|None = field(default=None)  # Where the gridded variable came from
    
    # Default attributes
    platform: str|None = field(default='platform')
    
    # Variable operation attributes
    to_grid: bool = field(default=False)  # If you want the variable to be gridded: True
    
    # Glider specific attributes
    id: str|None = field(default=None)
    wmo_id: str|None = field(default=None)
    instruments: str|None = field(default=None)
    type: str|None = field(default='platform')


    def __attrs_post_init__(self):
        """
        Post-initialization method to set the wmo_id attribute based on the id attribute.
        """
        if self.short_name is None and self.data_source_name is None:
            raise ValueError('Either the short_name or data_source_name attribute must be set.')
        # Generate update_time, long_name, and wmo_id if not given
        # Add current time to update_time if it was not given
        if self.update_time is None:
            self.update_time = pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')
        # Generate the long_name from the id if given
        if (self.long_name is None) & (self.id is not None):
            self.long_name = f'Slocum Glider {self.id}'
        # Generate the wmo_id from the id if given
        if (self.wmo_id is None) & (self.id is not None):
            self.wmo_id = get_wmo_id(self.id)
        if self.short_name is None:
            self.short_name = self.data_source_name
            
    def _filter_out_keys(self):
        """
        Filter out keys from the Variable object that are None.
        """
        # Return the dictionary sorted by key and filtered out None values  
        
        # Convert to_grid to string for JSON serialization
        self.to_grid = f'{self.to_grid}'
        self.data_source_name = str(self.data_source_name)
        
        return {key:value for key,value in asdict(self).items() if value is not None}
            
    def to_dict(self):
        """
        Convert the Variable object to a dictionary, sorted by key and filtered out None values.
        """
        # Return the dictionary sorted by key and filtered out None values
        return dict(sorted(self._filter_out_keys().items()))





from pathlib import Path

from glider_ingest import Processor
from glider_ingest.utils import timing

@timing
def main():
    """
    Example of how to use the MissionProcessor and MissionData classes to generate and save a mission dataset
    """    
    memory_card_copy_path = Path('C:/Users/alecmkrueger/Documents/GERG/GERG_GitHub/GERG-Glider/Code/Packages/glider_ingest/src/tests/test_data/memory_card_copy')
    # memory_card_copy_path = Path('G:/Shared drives/Slocum Gliders/Mission Data & Files/2024 Missions/Mission 48/Memory card copy')

    # Where you want the netcdf to be saved to
    working_dir = Path('C:/Users/alecmkrueger/Documents/GERG/GERG_GitHub/GERG-Glider/Code/Packages/glider_ingest/src/tests/test_data/working_dir').resolve()
    mission_num = '48'

    # Init a processor object
    processor = Processor(memory_card_copy_path=memory_card_copy_path,
                            working_dir=working_dir,
                            mission_num=mission_num)
    # Add custom variables to the mission_data container as a list of strings
    processor.add_mission_vars(mission_vars=['m_water_vy','m_water_vx'])
    # Save the container to a netcdf file
    processor.save()
    
    
if __name__ == '__main__':
    main()



from glider_ingest.variable import Variable
from glider_ingest.utils import get_polygon_coords,get_polygon_bounds

import numpy as np 
import pandas as pd
import uuid

# Flight Variables
def get_default_variables():
    m_pressure = Variable(
        data_source_name='m_pressure',
        short_name='m_pressure',
        accuracy=0.01,
        ancillary_variables='',
        axis='Z',
        bytes=4,
        comment='Alias for m_pressure',
        long_name='GPS Pressure',
        observation_type='measured',
        platform='platform',
        positive='down',
        precision=0.01,
        reference_datum='sea-surface',
        resolution=0.01,
        source_sensor='sci_water_pressure',
        standard_name='sea_water_pressure',
        units='bar',
        valid_max=2000.0,
        valid_min=0.0,
    )

    m_water_depth = Variable(
        data_source_name='m_water_depth',
        short_name='depth',
        accuracy=0.01,
        ancillary_variables='',
        axis='Z',
        bytes=4,
        comment='Alias for m_depth',
        long_name='GPS Depth',
        observation_type='calculated',
        platform='platform',
        positive='down',
        precision=0.01,
        reference_datum='sea-surface',
        resolution=0.01,
        source_sensor='m_depth',
        standard_name='sea_water_depth',
        units='meters',
        valid_max=2000.0,
        valid_min=0.0,
    )

    m_latitude = Variable(
        data_source_name='m_lat',
        short_name='latitude',
        ancillary_variables='',
        axis='Y',
        bytes=8,
        comment='m_gps_lat converted to decimal degrees and interpolated',
        coordinate_reference_frame='urn:ogc:crs:EPSG::4326',
        long_name='Latitude',
        observation_type='calculated',
        platform='platform',
        precision=5,
        reference_datum='WGS84',
        source_sensor='m_gps_lat',
        standard_name='latitude',
        units='degree_north',
        valid_max=90.0,
        valid_min=-90.0,
    )

    m_longitude = Variable(
        data_source_name='m_lon',
        short_name='longitude',
        ancillary_variables='',
        axis='X',
        bytes=8,
        comment='m_gps_lon converted to decimal degrees and interpolated',
        coordinate_reference_frame='urn:ogc:crs:EPSG::4326',
        long_name='Longitude',
        observation_type='calculated',
        platform='platform',
        precision=5,
        reference_datum='WGS84',
        source_sensor='m_gps_lon',
        standard_name='longitude',
        units='degree_east',
        valid_max=180.0,
        valid_min=-180.0,
    )

    # Science Variables

    sci_water_pressure = Variable(
        data_source_name='sci_water_pressure',
        short_name='pressure',
        accuracy=0.01,
        ancillary_variables='',
        axis='Z',
        bytes=4,
        comment='Alias for sci_water_pressure',
        instrument='instrument_ctd',
        long_name='CTD Pressure',
        observation_type='measured',
        platform='platform',
        positive='down',
        precision=0.01,
        reference_datum='sea-surface',
        resolution=0.01,
        source_sensor='sci_water_pressure',
        standard_name='sea_water_pressure',
        units='bar',
        valid_max=2000.0,
        valid_min=0.0,
    )

    sci_water_temp = Variable(
        data_source_name='sci_water_temp',
        short_name='temperature',
        accuracy=0.004,
        ancillary_variables='',
        bytes=4,
        instrument='instrument_ctd',
        long_name='Temperature',
        observation_type='measured',
        platform='platform',
        precision=0.001,
        resolution=0.001,
        standard_name='sea_water_temperature',
        units='Celsius',
        valid_max=40.0,
        valid_min=-5.0,
        to_grid=True
    )

    sci_water_cond = Variable(
        data_source_name='sci_water_cond',
        short_name='conductivity',
        accuracy=0.001,
        ancillary_variables='',
        bytes=4,
        instrument='instrument_ctd',
        long_name='sci_water_cond',
        observation_type='measured',
        platform='platform',
        precision=1e-05,
        resolution=1e-05,
        standard_name='sea_water_electrical_conductivity',
        units='S m-1',
        valid_max=10.0,
        valid_min=0.0,
        to_grid=True
    )

    sci_water_sal = Variable(
        data_source_name=None,
        short_name='salinity',
        accuracy='',
        ancillary_variables='',
        instrument='instrument_ctd',
        long_name='Salinity',
        observation_type='calculated',
        platform='platform',
        precision='',
        resolution='',
        standard_name='sea_water_practical_salinity',
        units='1',
        valid_max=40.0,
        valid_min=0.0,
        to_grid=True
    )

    sci_water_dens = Variable(
        data_source_name=None,
        short_name='density',
        accuracy='',
        ancillary_variables='',
        instrument='instrument_ctd',
        long_name='Density',
        observation_type='calculated',
        platform='platform',
        precision='',
        resolution='',
        standard_name='sea_water_density',
        units='kg m-3',
        valid_max=1040.0,
        valid_min=1015.0,
        to_grid=True
    )

    sci_flbbcd_bb_units = Variable(
        data_source_name='sci_flbbcd_bb_units',
        short_name='turbidity',
        accuracy='',
        ancillary_variables='',
        instrument='instrument_flbbcd',
        long_name='Turbidity',
        observation_type='calculated',
        platform='platform',
        precision='',
        resolution='',
        standard_name='sea_water_turbidity',
        units='1',
        valid_max=1.0,
        valid_min=0.0,
        to_grid=True
    )

    sci_flbbcd_cdom_units = Variable(
        data_source_name='sci_flbbcd_cdom_units',
        short_name='cdom',
        accuracy='',
        ancillary_variables='',
        instrument='instrument_flbbcd',
        long_name='CDOM',
        observation_type='calculated',
        platform='platform',
        precision='',
        resolution='',
        standard_name='concentration_of_colored_dissolved_organic_matter_in_sea_water',
        units='ppb',
        valid_max=50.0,
        valid_min=0.0,
        to_grid=True
    )

    sci_flbbcd_chlor_units = Variable(
        data_source_name='sci_flbbcd_chlor_units',
        short_name='chlorophyll',
        accuracy='',
        ancillary_variables='',
        instrument='instrument_flbbcd',
        long_name='Chlorophyll_a',
        observation_type='calculated',
        platform='platform',
        precision='',
        resolution='',
        standard_name='mass_concentration_of_chlorophyll_a_in_sea_water',
        units='\u03BCg/L',
        valid_max=10.0,
        valid_min=0.0,
        to_grid=True
    )

    sci_oxy4_oxygen = Variable(
        data_source_name='sci_oxy4_oxygen',
        short_name='oxygen',
        accuracy='',
        ancillary_variables='',
        instrument='instrument_ctd_modular_do_sensor',
        long_name='oxygen',
        observation_type='calculated',
        platform='platform',
        precision='',
        resolution='',
        standard_name='moles_of_oxygen_per_unit_mass_in_sea_water',
        units='\u03BCmol/kg',
        valid_max=500.0,
        valid_min=0.0,
        to_grid=True
    )
    
    default_variables = [
        m_pressure,
        m_water_depth,
        m_latitude,
        m_longitude,
        sci_water_pressure,
        sci_water_temp,
        sci_water_cond,
        sci_water_sal,
        sci_water_dens,
        sci_flbbcd_bb_units,
        sci_flbbcd_cdom_units,
        sci_flbbcd_chlor_units,
        sci_oxy4_oxygen
    ]
    
    return default_variables

def get_global_attrs(wmo_id:str,mission_title:str,longitude:np.ndarray,latitude:np.ndarray,depth:np.ndarray,time:np.ndarray):
    
    # Calculate spatial bounds and resolution
    lat_max, lat_min, lon_max, lon_min = get_polygon_bounds(latitude=latitude, longitude=longitude)
    geospatial_bounds = get_polygon_coords(longitude=longitude, latitude=latitude,
                                           lat_max=lat_max, lat_min=lat_min, lon_max=lon_max, lon_min=lon_min)
    
    geospatial_lat_resolution = "{:.4e}".format(abs(np.nanmean(np.diff(latitude))))+ ' degree'
    geospatial_lon_resolution = "{:.4e}".format(abs(np.nanmean(np.diff(longitude))))+ ' degree'
    
    vertical_min = np.nanmin(depth[np.where(depth>0)])
    vertical_max = np.nanmax(depth)
    
    # Get current time
    current_time = pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')
    # Get dataset time range
    time_coverage_start = time[-1]
    time_coverage_end = time[-1]
    time_coverage_duration = f"PT{str((time_coverage_end - time_coverage_start) / np.timedelta64(1, 's'))}S"
    
    # Get uuid
    uuid_str = str(uuid.uuid4())
    
    
    global_attrs = {'Conventions': 'CF-1.6, COARDS, ACDD-1.3',
                    'acknowledgment': ' ',
                    'cdm_data_type': 'Profile',
                    'comment': 'time is the ctd_time from sci_m_present_time, g_time and g_pres are the grided time and pressure',
                    'contributor_name': 'Steven F. DiMarco',
                    'contributor_role': ' ',
                    'creator_email': 'sakib@tamu.edu, gexiao@tamu.edu',
                    'creator_institution': 'Texas A&M University, Geochemical and Environmental Research Group',
                    'creator_name': 'Sakib Mahmud, Xiao Ge',
                    'creator_type': 'persons',
                    'creator_url': 'https://gerg.tamu.edu/',
                    'date_created': current_time,
                    'date_issued': current_time,
                    'date_metadata_modified': '2023-09-15',
                    'date_modified': current_time,
                    'deployment': ' ',
                    'featureType': 'profile',
                    'geospatial_bounds_crs': 'EPSG:4326',
                    'geospatial_bounds_vertical_crs': 'EPSG:5831',
                    'geospatial_lat_resolution': geospatial_lat_resolution,
                    'geospatial_lat_units': 'degree_north',
                    'geospatial_lon_resolution': geospatial_lon_resolution,
                    'geospatial_lon_units': 'degree_east',
                    'geospatial_vertical_positive': 'down',
                    'geospatial_vertical_resolution': ' ',
                    'geospatial_vertical_units': 'EPSG:5831',
                    'infoUrl': 'https://gerg.tamu.edu/',
                    'institution': 'Texas A&M University, Geochemical and Environmental Research Group',
                    'instrument': 'In Situ/Laboratory Instruments > Profilers/Sounders > CTD',
                    'instrument_vocabulary': 'NASA/GCMD Instrument Keywords Version 8.5',
                    'ioos_regional_association': 'GCOOS-RA',
                    'keywords': 'Oceans > Ocean Pressure > Water Pressure, Oceans > Ocean Temperature > Water Temperature, Oceans > Salinity/Density > Conductivity, Oceans > Salinity/Density > Density, Oceans > Salinity/Density > Salinity',
                    'keywords_vocabulary': 'NASA/GCMD Earth Sciences Keywords Version 8.5',
                    'license': 'This data may be redistributed and used without restriction.  Data provided as is with no expressed or implied assurance of quality assurance or quality control',
                    'metadata_link': ' ',
                    'naming_authority': 'org.gcoos.gandalf',
                    'ncei_template_version': 'NCEI_NetCDF_Trajectory_Template_v2.0',
                    'platform': 'In Situ Ocean-based Platforms > AUVS > Autonomous Underwater Vehicles',
                    'platform_type': 'Slocum Glider',
                    'platform_vocabulary': 'NASA/GCMD Platforms Keywords Version 8.5',
                    'processing_level': 'Level 0',
                    'product_version': '0.0',
                    'program': ' ',
                    'project': ' ',
                    'publisher_email': 'sdimarco@tamu.edu',
                    'publisher_institution': 'Texas A&M University, Geochemical and Environmental Research Group',
                    'publisher_name': 'Steven F. DiMarco',
                    'publisher_url': 'https://gerg.tamu.edu/',
                    'references': ' ',
                    'sea_name': 'Gulf of Mexico',
                    'standard_name_vocabulary': 'CF Standard Name Table v27',
                    'summary': 'Merged dataset for GERG future usage.',
                    'time_coverage_resolution': ' ',
                    'wmo_id': wmo_id,
                    'uuid': uuid_str,
                    'history': 'dbd and ebd files transferred from dbd2asc on 2023-09-15, merged into single netCDF file on '+current_time,
                    'title': mission_title,
                    'source': 'Observational Slocum glider data from source ebd and dbd files',
                    'geospatial_lat_min': str(lat_min),
                    'geospatial_lat_max': str(lat_max),
                    'geospatial_lon_min': str(lon_min),
                    'geospatial_lon_max': str(lon_max),
                    'geospatial_bounds': geospatial_bounds,
                    'geospatial_vertical_min': str(vertical_min),
                    'geospatial_vertical_max': str(vertical_max),
                    'time_coverage_start': str(time_coverage_start)[:19],
                    'time_coverage_end': str(time_coverage_end)[:19],
                    'time_coverage_duration': time_coverage_duration}
    
    return global_attrs